import os
import time
import pg8000
import datetime
import pytest


@pytest.fixture
def has_tzset():

    # Neither Windows nor Jython 2.5.3 have a time.tzset() so skip
    if hasattr(time, 'tzset'):
        os.environ['TZ'] = "UTC"
        time.tzset()
        return True
    return False


# DBAPI compatible interface tests
@pytest.fixture
def db_table(con, has_tzset):
    with con.cursor() as c:
        c.execute(
            "CREATE TEMPORARY TABLE t1 "
            "(f1 int primary key, f2 int not null, f3 varchar(50) null) "
            "ON COMMIT DROP")
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (%s, %s, %s)", (1, 1, None))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (%s, %s, %s)", (2, 10, None))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (%s, %s, %s)", (3, 100, None))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (%s, %s, %s)", (4, 1000, None))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (%s, %s, %s)",
            (5, 10000, None))
    return con


def test_parallel_queries(db_table):
    with db_table.cursor() as c1, db_table.cursor() as c2:

        c1.execute("SELECT f1, f2, f3 FROM t1")
        while 1:
            row = c1.fetchone()
            if row is None:
                break
            f1, f2, f3 = row
            c2.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > %s", (f1,))
            while 1:
                row = c2.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row


def test_qmark(db_table):
    orig_paramstyle = pg8000.paramstyle
    try:
        pg8000.paramstyle = "qmark"
        with db_table.cursor() as c1:
            c1.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > ?", (3,))
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        pg8000.paramstyle = orig_paramstyle


def test_numeric(db_table):
    orig_paramstyle = pg8000.paramstyle
    try:
        pg8000.paramstyle = "numeric"
        with db_table.cursor() as c1:
            c1.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > :1", (3,))
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        pg8000.paramstyle = orig_paramstyle


def test_named(db_table):
    orig_paramstyle = pg8000.paramstyle
    try:
        pg8000.paramstyle = "named"
        with db_table.cursor() as c1:
            c1.execute(
                "SELECT f1, f2, f3 FROM t1 WHERE f1 > :f1", {"f1": 3})
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        pg8000.paramstyle = orig_paramstyle


def test_format(db_table):
    orig_paramstyle = pg8000.paramstyle
    try:
        pg8000.paramstyle = "format"
        with db_table.cursor() as c1:
            c1.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > %s", (3,))
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        pg8000.paramstyle = orig_paramstyle


def test_pyformat(db_table):
    orig_paramstyle = pg8000.paramstyle
    try:
        pg8000.paramstyle = "pyformat"
        with db_table.cursor() as c1:
            c1.execute(
                "SELECT f1, f2, f3 FROM t1 WHERE f1 > %(f1)s", {"f1": 3})
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        pg8000.paramstyle = orig_paramstyle


def test_arraysize(db_table):
    with db_table.cursor() as c1:
        c1.arraysize = 3
        c1.execute("SELECT * FROM t1")
        retval = c1.fetchmany()
        assert len(retval) == c1.arraysize


def test_date():
    val = pg8000.Date(2001, 2, 3)
    assert val == datetime.date(2001, 2, 3)


def test_time():
    val = pg8000.Time(4, 5, 6)
    assert val == datetime.time(4, 5, 6)


def test_timestamp():
    val = pg8000.Timestamp(2001, 2, 3, 4, 5, 6)
    assert val == datetime.datetime(2001, 2, 3, 4, 5, 6)


def test_date_from_ticks(has_tzset):
    if has_tzset:
        val = pg8000.DateFromTicks(1173804319)
        assert val == datetime.date(2007, 3, 13)


def testTimeFromTicks(has_tzset):
    if has_tzset:
        val = pg8000.TimeFromTicks(1173804319)
        assert val == datetime.time(16, 45, 19)


def test_timestamp_from_ticks(has_tzset):
    if has_tzset:
        val = pg8000.TimestampFromTicks(1173804319)
        assert val == datetime.datetime(2007, 3, 13, 16, 45, 19)


def test_binary():
    v = pg8000.Binary(b"\x00\x01\x02\x03\x02\x01\x00")
    assert v == b"\x00\x01\x02\x03\x02\x01\x00"
    assert isinstance(v, pg8000.BINARY)


def test_row_count(db_table):
    with db_table.cursor() as c1:
        c1.execute("SELECT * FROM t1")

        assert 5 == c1.rowcount

        c1.execute("UPDATE t1 SET f3 = %s WHERE f2 > 101", ("Hello!",))
        assert 2 == c1.rowcount

        c1.execute("DELETE FROM t1")
        assert 5 == c1.rowcount


def test_fetch_many(db_table):
    with db_table.cursor() as cursor:
        cursor.arraysize = 2
        cursor.execute("SELECT * FROM t1")
        assert 2 == len(cursor.fetchmany())
        assert 2 == len(cursor.fetchmany())
        assert 1 == len(cursor.fetchmany())
        assert 0 == len(cursor.fetchmany())


def test_iterator(db_table):
    with db_table.cursor() as cursor:
        cursor.execute("SELECT * FROM t1 ORDER BY f1")
        f1 = 0
        for row in cursor:
            next_f1 = row[0]
            assert next_f1 > f1
            f1 = next_f1


# Vacuum can't be run inside a transaction, so we need to turn
# autocommit on.
def test_vacuum(con):
    con.autocommit = True
    with con.cursor() as cursor:
        cursor.execute("vacuum")


def test_prepared_statement(con):
    with con.cursor() as cursor:
        cursor.execute('PREPARE gen_series AS SELECT generate_series(1, 10);')
        cursor.execute('EXECUTE gen_series')


def test_cursor_type(cursor):
    assert str(type(cursor)) == "<class 'pg8000.core.Cursor'>"



import pg8000
import sys
import socket
import struct
import pytest
import ssl


# Check if running in Jython
if 'java' in sys.platform:
    from javax.net.ssl import TrustManager, X509TrustManager
    from jarray import array
    from javax.net.ssl import SSLContext

    class TrustAllX509TrustManager(X509TrustManager):
        '''Define a custom TrustManager which will blindly accept all
        certificates'''

        def checkClientTrusted(self, chain, auth):
            pass

        def checkServerTrusted(self, chain, auth):
            pass

        def getAcceptedIssuers(self):
            return None
    # Create a static reference to an SSLContext which will use
    # our custom TrustManager
    trust_managers = array([TrustAllX509TrustManager()], TrustManager)
    TRUST_ALL_CONTEXT = SSLContext.getInstance("SSL")
    TRUST_ALL_CONTEXT.init(None, trust_managers, None)
    # Keep a static reference to the JVM's default SSLContext for restoring
    # at a later time
    DEFAULT_CONTEXT = SSLContext.getDefault()


@pytest.fixture
def trust_all_certificates(request):
    '''Decorator function that will make it so the context of the decorated
    method will run with our TrustManager that accepts all certificates'''
    # Only do this if running under Jython
    is_java = 'java' in sys.platform

    if is_java:
        from javax.net.ssl import SSLContext
        SSLContext.setDefault(TRUST_ALL_CONTEXT)

    def fin():
        if is_java:
            SSLContext.setDefault(DEFAULT_CONTEXT)

    request.addfinalizer(fin)


def testSocketMissing():
    conn_params = {
        'unix_sock': "/file-does-not-exist",
        'user': "doesn't-matter"}

    with pytest.raises(pg8000.InterfaceError):
        pg8000.connect(**conn_params)


def testDatabaseMissing(db_kwargs):
    db_kwargs["database"] = "missing-db"
    with pytest.raises(pg8000.ProgrammingError):
        pg8000.connect(**db_kwargs)


def testNotify(con):
    assert list(con.notifications) == []
    with con.cursor() as cursor:
        cursor.execute("LISTEN test")
        cursor.execute("NOTIFY test")
        con.commit()

        cursor.execute("VALUES (1, 2), (3, 4), (5, 6)")
        assert len(con.notifications) == 1
        assert con.notifications[0][1] == "test"


# This requires a line in pg_hba.conf that requires md5 for the database
# pg8000_md5

def testMd5(db_kwargs):
    db_kwargs["database"] = "pg8000_md5"

    # Should only raise an exception saying db doesn't exist
    with pytest.raises(pg8000.ProgrammingError, match='3D000'):
        pg8000.connect(**db_kwargs)


# This requires a line in pg_hba.conf that requires gss for the database
# pg8000_gss

def testGss(db_kwargs):
    db_kwargs["database"] = "pg8000_gss"

    # Should raise an exception saying gss isn't supported
    with pytest.raises(
            pg8000.InterfaceError,
            match="Authentication method 7 not supported by pg8000."):
        pg8000.connect(**db_kwargs)


@pytest.mark.usefixtures("trust_all_certificates")
def testSsl(db_kwargs):
    context = ssl.SSLContext()
    context.check_hostname = False
    db_kwargs["ssl_context"] = context
    with pg8000.connect(**db_kwargs):
        pass


# This requires a line in pg_hba.conf that requires 'password' for the
# database pg8000_password

def testPassword(db_kwargs):
    db_kwargs["database"] = "pg8000_password"

    # Should only raise an exception saying db doesn't exist
    with pytest.raises(pg8000.ProgrammingError, match='3D000'):
        pg8000.connect(**db_kwargs)


def testUnicodeDatabaseName(db_kwargs):
    db_kwargs["database"] = "pg8000_sn\uFF6Fw"

    # Should only raise an exception saying db doesn't exist
    with pytest.raises(pg8000.ProgrammingError, match='3D000'):
        pg8000.connect(**db_kwargs)


def testBytesDatabaseName(db_kwargs):
    """ Should only raise an exception saying db doesn't exist """

    db_kwargs["database"] = bytes("pg8000_sn\uFF6Fw", 'utf8')
    with pytest.raises(pg8000.ProgrammingError, match='3D000'):
        pg8000.connect(**db_kwargs)


def testBytesPassword(con, db_kwargs):
    # Create user
    username = 'boltzmann'
    password = 'cha\uFF6Fs'
    with con.cursor() as cur:
        cur.execute(
            "create user " + username + " with password '" + password + "';")
        con.commit()

        db_kwargs['user'] = username
        db_kwargs['password'] = password.encode('utf8')
        db_kwargs['database'] = 'pg8000_md5'
        with pytest.raises(pg8000.ProgrammingError, match='3D000'):
            pg8000.connect(**db_kwargs)

        cur.execute("drop role " + username)
        con.commit()


def test_broken_pipe(con, db_kwargs):
    with pg8000.connect(**db_kwargs) as db1:
        with db1.cursor() as cur1, con.cursor() as cur2:
            cur1.execute("select pg_backend_pid()")
            pid1 = cur1.fetchone()[0]

            cur2.execute("select pg_terminate_backend(%s)", (pid1,))
            try:
                cur1.execute("select 1")
            except Exception as e:
                assert isinstance(e, (socket.error, struct.error))


def testApplicatioName(db_kwargs):
    app_name = 'my test application name'
    db_kwargs['application_name'] = app_name
    with pg8000.connect(**db_kwargs) as db:
        cur = db.cursor()
        cur.execute(
            'select application_name from pg_stat_activity '
            ' where pid = pg_backend_pid()')

        application_name = cur.fetchone()[0]
        assert application_name == app_name


def test_application_name_integer(db_kwargs):
    db_kwargs['application_name'] = 1
    with pytest.raises(
            pg8000.InterfaceError,
            match="The parameter application_name can't be of type "
            "<class 'int'>."):
        pg8000.connect(**db_kwargs)


def test_application_name_bytearray(db_kwargs):
    db_kwargs['application_name'] = bytearray(b'Philby')
    pg8000.connect(**db_kwargs)


# This requires a line in pg_hba.conf that requires scram-sha-256 for the
# database scram-sha-256

def test_scram_sha_256(db_kwargs):
    db_kwargs["database"] = "pg8000_scram_sha_256"

    # Should only raise an exception saying db doesn't exist
    with pytest.raises(pg8000.ProgrammingError, match='3D000'):
        pg8000.connect(**db_kwargs)



import pg8000
from pg8000 import PGJsonb, PGEnum
from datetime import (
    datetime as Datetime, time as Time, date as Date, timedelta as Timedelta,
    timezone as Timezone)
import decimal
import struct
import uuid
import os
import time
import json
import pytz
from collections import OrderedDict
import pytest
from enum import Enum
import ipaddress


# Type conversion tests

def testTimeRoundtrip(cursor):
    cursor.execute("SELECT %s as f1", (Time(4, 5, 6),))
    retval = cursor.fetchall()
    assert retval[0][0] == Time(4, 5, 6)


def test_date_roundtrip(cursor):
    v = Date(2001, 2, 3)
    cursor.execute("SELECT %s as f1", (v,))
    retval = cursor.fetchall()
    assert retval[0][0] == v


def test_bool_roundtrip(cursor):
    cursor.execute("SELECT %s as f1", (True,))
    retval = cursor.fetchall()
    assert retval[0][0] is True


def test_null_roundtrip(cursor):
    cursor.execute("show server_version_num")
    version = cursor.fetchone()[0][:2]

    if version.startswith('9'):
        # Prior to PostgreSQL version 10 We can't just "SELECT %s" and set
        # None as the parameter, since it has no type.  That would result
        # in a PG error, "could not determine data type of parameter %s".
        # So we create a temporary table, insert null values, and read them
        # back.
        cursor.execute(
            "CREATE TEMPORARY TABLE TestNullWrite "
            "(f1 int4, f2 timestamp, f3 varchar)")
        cursor.execute(
            "INSERT INTO TestNullWrite VALUES (%s, %s, %s)",
            (None, None, None,))
        cursor.execute("SELECT * FROM TestNullWrite")
        retval = cursor.fetchone()
        assert retval == [None, None, None]

        with pytest.raises(pg8000.ProgrammingError):
            cursor.execute("SELECT %s as f1", (None,))
    else:
        cursor.execute("SELECT %s as f1", (None,))
        assert cursor.fetchone()[0] is None


def test_decimal_roundtrip(cursor):
    values = (
        "1.1", "-1.1", "10000", "20000", "-1000000000.123456789", "1.0",
        "12.44")
    for v in values:
        cursor.execute("SELECT %s as f1", (decimal.Decimal(v),))
        retval = cursor.fetchall()
        assert str(retval[0][0]) == v


def test_float_roundtrip(cursor):
    # This test ensures that the binary float value doesn't change in a
    # roundtrip to the server.  That could happen if the value was
    # converted to text and got rounded by a decimal place somewhere.
    val = 1.756e-12
    bin_orig = struct.pack("!d", val)
    cursor.execute("SELECT %s as f1", (val,))
    retval = cursor.fetchall()
    bin_new = struct.pack("!d", retval[0][0])
    assert bin_new == bin_orig


def test_float_plus_infinity_roundtrip(cursor):
    v = float('inf')
    cursor.execute("SELECT %s as f1", (v,))
    retval = cursor.fetchall()
    assert retval[0][0] == v


def test_str_roundtrip(cursor):
    v = "hello world"
    cursor.execute(
        "create temporary table test_str (f character varying(255))")
    cursor.execute("INSERT INTO test_str VALUES (%s)", (v,))
    retval = tuple(cursor.execute("SELECT * from test_str"))
    assert retval[0][0] == v


def test_str_then_int(cursor):
    v1 = "hello world"
    retval = tuple(cursor.execute("SELECT cast(%s as varchar) as f1", (v1,)))
    assert retval[0][0] == v1

    v2 = 1
    retval = tuple(cursor.execute("SELECT cast(%s as varchar) as f1", (v2,)))
    assert retval[0][0] == str(v2)


def test_unicode_roundtrip(cursor):
    v = "hello \u0173 world"
    retval = tuple(cursor.execute("SELECT cast(%s as varchar) as f1", (v,)))
    assert retval[0][0] == v


def test_long_roundtrip(cursor):
    v = 50000000000000
    retval = tuple(cursor.execute("SELECT %s", (v,)))
    assert retval[0][0] == v


def test_int_execute_many(cursor):
    tuple(cursor.executemany("SELECT %s", ((1,), (40000,))))

    v = ([None], [4])
    cursor.execute("create temporary table test_int (f integer)")
    cursor.executemany("INSERT INTO test_int VALUES (%s)", v)
    retval = tuple(cursor.execute("SELECT * from test_int"))
    assert retval == v


def test_int_roundtrip(cursor):
    int2 = 21
    int4 = 23
    int8 = 20

    test_values = [
        (0, int2),
        (-32767, int2),
        (-32768, int4),
        (+32767, int2),
        (+32768, int4),
        (-2147483647, int4),
        (-2147483648, int8),
        (+2147483647, int4),
        (+2147483648, int8),
        (-9223372036854775807, int8),
        (+9223372036854775807, int8)]

    for value, typoid in test_values:
        cursor.execute("SELECT %s", (value,))
        retval = cursor.fetchall()
        assert retval[0][0] == value
        column_name, column_typeoid = cursor.description[0][0:2]
        assert column_typeoid == typoid


def test_bytea_roundtrip(cursor):
    cursor.execute(
        "SELECT %s as f1", (pg8000.Binary(b"\x00\x01\x02\x03\x02\x01\x00"),))
    retval = cursor.fetchall()
    assert retval[0][0] == b"\x00\x01\x02\x03\x02\x01\x00"


def test_bytearray_round_trip(cursor):
    binary = b'\x00\x01\x02\x03\x02\x01\x00'
    cursor.execute("SELECT %s as f1", (bytearray(binary),))
    retval = cursor.fetchall()
    assert retval[0][0] == binary


def test_bytearray_subclass_round_trip(cursor):
    class BClass(bytearray):
        pass
    binary = b'\x00\x01\x02\x03\x02\x01\x00'
    cursor.execute("SELECT %s as f1", (BClass(binary),))
    retval = cursor.fetchall()
    assert retval[0][0] == binary


def test_timestamp_roundtrip(is_java, cursor):
    v = Datetime(2001, 2, 3, 4, 5, 6, 170000)
    cursor.execute("SELECT %s as f1", (v,))
    retval = cursor.fetchall()
    assert retval[0][0] == v

    # Test that time zone doesn't affect it
    # Jython 2.5.3 doesn't have a time.tzset() so skip
    if not is_java:
        orig_tz = os.environ.get('TZ')
        os.environ['TZ'] = "America/Edmonton"
        time.tzset()

        cursor.execute("SELECT %s as f1", (v,))
        retval = cursor.fetchall()
        assert retval[0][0] == v

        if orig_tz is None:
            del os.environ['TZ']
        else:
            os.environ['TZ'] = orig_tz
        time.tzset()


def test_interval_roundtrip(cursor):
    v = pg8000.Interval(microseconds=123456789, days=2, months=24)
    cursor.execute("SELECT %s as f1", (v,))
    retval = cursor.fetchall()
    assert retval[0][0] == v

    v = Timedelta(seconds=30)
    cursor.execute("SELECT %s as f1", (v,))
    retval = cursor.fetchall()
    assert retval[0][0] == v


def test_enum_str_round_trip(cursor):
    try:
        cursor.execute(
            "create type lepton as enum ('electron', 'muon', 'tau')")

        v = 'muon'
        cursor.execute("SELECT cast(%s as lepton) as f1", (v,))
        retval = cursor.fetchall()
        assert retval[0][0] == v
        cursor.execute("CREATE TEMPORARY TABLE testenum (f1 lepton)")
        cursor.execute(
            "INSERT INTO testenum VALUES (cast(%s as lepton))", ('electron',))
    finally:
        cursor.execute("drop table testenum")
        cursor.execute("drop type lepton")


def test_enum_custom_round_trip(cursor):
    class Lepton(object):
        # Implements PEP 435 in the minimal fashion needed
        __members__ = OrderedDict()

        def __init__(self, name, value, alias=None):
            self.name = name
            self.value = value
            self.__members__[name] = self
            setattr(self.__class__, name, self)
            if alias:
                self.__members__[alias] = self
                setattr(self.__class__, alias, self)

    try:
        cursor.execute("create type lepton as enum ('1', '2', '3')")

        v = Lepton('muon', '2')
        cursor.execute("SELECT cast(%s as lepton) as f1", (PGEnum(v),))
        retval = cursor.fetchall()
        assert retval[0][0] == v.value
    finally:
        cursor.execute("drop type lepton")


def test_enum_py_round_trip(cursor):
    class Lepton(Enum):
        electron = '1'
        muon = '2'
        tau = '3'

    try:
        cursor.execute("create type lepton as enum ('1', '2', '3')")

        v = Lepton.muon
        retval = tuple(cursor.execute("SELECT cast(%s as lepton) as f1", (v,)))
        assert retval[0][0] == v.value

        cursor.execute("CREATE TEMPORARY TABLE testenum (f1 lepton)")
        cursor.execute(
            "INSERT INTO testenum VALUES (cast(%s as lepton))",
            (Lepton.electron,))
    finally:
        cursor.execute("drop table testenum")
        cursor.execute("drop type lepton")


def test_xml_roundtrip(cursor):
    v = '<genome>gatccgagtac</genome>'
    retval = tuple(cursor.execute("select xmlparse(content %s) as f1", (v,)))
    assert retval[0][0] == v


def test_uuid_roundtrip(cursor):
    v = uuid.UUID('911460f2-1f43-fea2-3e2c-e01fd5b5069d')
    retval = tuple(cursor.execute("select %s as f1", (v,)))
    assert retval[0][0] == v


def test_inet_roundtrip(cursor):
    v = ipaddress.ip_network('192.168.0.0/28')
    retval = tuple(cursor.execute("select %s as f1", (v,)))
    assert retval[0][0] == v

    v = ipaddress.ip_address('192.168.0.1')
    retval = tuple(cursor.execute("select %s as f1", (v,)))
    assert retval[0][0] == v


def test_xid_roundtrip(cursor):
    v = 86722
    cursor.execute("select cast(cast(%s as varchar) as xid) as f1", (v,))
    retval = cursor.fetchall()
    assert retval[0][0] == v

    # Should complete without an exception
    cursor.execute("select * from pg_locks where transactionid = %s", (97712,))
    retval = cursor.fetchall()


def test_int2vector_in(cursor):
    retval = tuple(cursor.execute("select cast('1 2' as int2vector) as f1"))
    assert retval[0][0] == [1, 2]

    # Should complete without an exception
    tuple(cursor.execute("select indkey from pg_index"))


def test_timestamp_tz_out(cursor):
    cursor.execute(
        "SELECT '2001-02-03 04:05:06.17 America/Edmonton'"
        "::timestamp with time zone")
    retval = cursor.fetchall()
    dt = retval[0][0]
    assert dt.tzinfo is not None, "no tzinfo returned"
    assert dt.astimezone(Timezone.utc) == Datetime(
        2001, 2, 3, 11, 5, 6, 170000, Timezone.utc), \
        "retrieved value match failed"


def test_timestamp_tz_roundtrip(is_java, cursor):
    if not is_java:
        mst = pytz.timezone("America/Edmonton")
        v1 = mst.localize(Datetime(2001, 2, 3, 4, 5, 6, 170000))
        cursor.execute("SELECT %s as f1", (v1,))
        retval = cursor.fetchall()
        v2 = retval[0][0]
        assert v2.tzinfo is not None
        assert v1 == v2


def test_timestamp_mismatch(is_java, cursor):
    if not is_java:
        mst = pytz.timezone("America/Edmonton")
        cursor.execute("SET SESSION TIME ZONE 'America/Edmonton'")
        try:
            cursor.execute(
                "CREATE TEMPORARY TABLE TestTz "
                "(f1 timestamp with time zone, "
                "f2 timestamp without time zone)")
            cursor.execute(
                "INSERT INTO TestTz (f1, f2) VALUES (%s, %s)", (
                    # insert timestamp into timestamptz field (v1)
                    Datetime(2001, 2, 3, 4, 5, 6, 170000),
                    # insert timestamptz into timestamp field (v2)
                    mst.localize(Datetime(
                        2001, 2, 3, 4, 5, 6, 170000))))
            cursor.execute("SELECT f1, f2 FROM TestTz")
            retval = cursor.fetchall()

            # when inserting a timestamp into a timestamptz field,
            # postgresql assumes that it is in local time. So the value
            # that comes out will be the server's local time interpretation
            # of v1. We've set the server's TZ to MST, the time should
            # be...
            f1 = retval[0][0]
            assert f1 == Datetime(2001, 2, 3, 11, 5, 6, 170000, Timezone.utc)

            # inserting the timestamptz into a timestamp field, pg8000
            # converts the value into UTC, and then the PG server converts
            # it into local time for insertion into the field. When we
            # query for it, we get the same time back, like the tz was
            # dropped.
            f2 = retval[0][1]
            assert f2 == Datetime(2001, 2, 3, 4, 5, 6, 170000)
        finally:
            cursor.execute("SET SESSION TIME ZONE DEFAULT")


def test_name_out(cursor):
    # select a field that is of "name" type:
    tuple(cursor.execute("SELECT usename FROM pg_user"))
    # It is sufficient that no errors were encountered.


def test_oid_out(cursor):
    tuple(cursor.execute("SELECT oid FROM pg_type"))
    # It is sufficient that no errors were encountered.


def test_boolean_out(cursor):
    retval = tuple(cursor.execute("SELECT cast('t' as bool)"))
    assert retval[0][0]


def test_numeric_out(cursor):
    for num in ('5000', '50.34'):
        retval = tuple(cursor.execute("SELECT " + num + "::numeric"))
        assert str(retval[0][0]) == num


def test_int2_out(cursor):
    retval = tuple(cursor.execute("SELECT 5000::smallint"))
    assert retval[0][0] == 5000


def test_int4_out(cursor):
    retval = tuple(cursor.execute("SELECT 5000::integer"))
    assert retval[0][0] == 5000


def test_int8_out(cursor):
    retval = tuple(cursor.execute("SELECT 50000000000000::bigint"))
    assert retval[0][0] == 50000000000000


def test_float4_out(cursor):
    retval = tuple(cursor.execute("SELECT 1.1::real"))
    assert retval[0][0] == 1.1000000238418579


def test_float8_out(cursor):
    retval = tuple(cursor.execute("SELECT 1.1::double precision"))
    assert retval[0][0] == 1.1000000000000001


def test_varchar_out(cursor):
    retval = tuple(cursor.execute("SELECT 'hello'::varchar(20)"))
    assert retval[0][0] == "hello"


def test_char_out(cursor):
    retval = tuple(cursor.execute("SELECT 'hello'::char(20)"))
    assert retval[0][0] == "hello               "


def test_text_out(cursor):
    retval = tuple(cursor.execute("SELECT 'hello'::text"))
    assert retval[0][0] == "hello"


def test_interval_out(cursor):
    retval = tuple(
        cursor.execute(
            "SELECT '1 month 16 days 12 hours 32 minutes 64 seconds'"
            "::interval"))
    expected_value = pg8000.Interval(
        microseconds=(12 * 60 * 60 * 1000 * 1000) +
        (32 * 60 * 1000 * 1000) + (64 * 1000 * 1000), days=16, months=1)
    assert retval[0][0] == expected_value

    retval = tuple(cursor.execute("select interval '30 seconds'"))
    assert retval[0][0] == Timedelta(seconds=30)

    retval = tuple(cursor.execute("select interval '12 days 30 seconds'"))
    assert retval[0][0] == Timedelta(days=12, seconds=30)


def test_timestamp_out(cursor):
    cursor.execute("SELECT '2001-02-03 04:05:06.17'::timestamp")
    retval = cursor.fetchall()
    assert retval[0][0] == Datetime(2001, 2, 3, 4, 5, 6, 170000)


# confirms that pg8000's binary output methods have the same output for
# a data type as the PG server
def test_binary_output_methods(con):
    with con.cursor() as cursor:
        methods = (
            ("float8send", 22.2),
            ("timestamp_send", Datetime(2001, 2, 3, 4, 5, 6, 789)),
            ("byteasend", pg8000.Binary(b"\x01\x02")),
            ("interval_send", pg8000.Interval(1234567, 123, 123)),)
        for method_out, value in methods:
            cursor.execute("SELECT %s(%%s) as f1" % method_out, (value,))
            retval = cursor.fetchall()
            assert retval[0][0] == con.make_params((value,))[0][2](value)


def test_int4_array_out(cursor):
    cursor.execute(
        "SELECT '{1,2,3,4}'::INT[] AS f1, '{{1,2,3},{4,5,6}}'::INT[][] AS f2, "
        "'{{{1,2},{3,4}},{{NULL,6},{7,8}}}'::INT[][][] AS f3")
    f1, f2, f3 = cursor.fetchone()
    assert f1 == [1, 2, 3, 4]
    assert f2 == [[1, 2, 3], [4, 5, 6]]
    assert f3 == [[[1, 2], [3, 4]], [[None, 6], [7, 8]]]


def test_int2_array_out(cursor):
    cursor.execute(
        "SELECT '{1,2,3,4}'::INT2[] AS f1, "
        "'{{1,2,3},{4,5,6}}'::INT2[][] AS f2, "
        "'{{{1,2},{3,4}},{{NULL,6},{7,8}}}'::INT2[][][] AS f3")
    f1, f2, f3 = cursor.fetchone()
    assert f1 == [1, 2, 3, 4]
    assert f2 == [[1, 2, 3], [4, 5, 6]]
    assert f3 == [[[1, 2], [3, 4]], [[None, 6], [7, 8]]]


def test_int8_array_out(cursor):
    cursor.execute(
        "SELECT '{1,2,3,4}'::INT8[] AS f1, "
        "'{{1,2,3},{4,5,6}}'::INT8[][] AS f2, "
        "'{{{1,2},{3,4}},{{NULL,6},{7,8}}}'::INT8[][][] AS f3")
    f1, f2, f3 = cursor.fetchone()
    assert f1 == [1, 2, 3, 4]
    assert f2 == [[1, 2, 3], [4, 5, 6]]
    assert f3 == [[[1, 2], [3, 4]], [[None, 6], [7, 8]]]


def test_bool_array_out(cursor):
    cursor.execute(
        "SELECT '{TRUE,FALSE,FALSE,TRUE}'::BOOL[] AS f1, "
        "'{{TRUE,FALSE,TRUE},{FALSE,TRUE,FALSE}}'::BOOL[][] AS f2, "
        "'{{{TRUE,FALSE},{FALSE,TRUE}},{{NULL,TRUE},{FALSE,FALSE}}}'"
        "::BOOL[][][] AS f3")
    f1, f2, f3 = cursor.fetchone()
    assert f1 == [True, False, False, True]
    assert f2 == [[True, False, True], [False, True, False]]
    assert f3 == [
        [[True, False], [False, True]], [[None, True], [False, False]]]


def test_float4_array_out(cursor):
    cursor.execute(
        "SELECT '{1,2,3,4}'::FLOAT4[] AS f1, "
        "'{{1,2,3},{4,5,6}}'::FLOAT4[][] AS f2, "
        "'{{{1,2},{3,4}},{{NULL,6},{7,8}}}'::FLOAT4[][][] AS f3")
    f1, f2, f3 = cursor.fetchone()
    assert f1 == [1, 2, 3, 4]
    assert f2 == [[1, 2, 3], [4, 5, 6]]
    assert f3 == [[[1, 2], [3, 4]], [[None, 6], [7, 8]]]


def test_float8_array_out(cursor):
    cursor.execute(
        "SELECT '{1,2,3,4}'::FLOAT8[] AS f1, "
        "'{{1,2,3},{4,5,6}}'::FLOAT8[][] AS f2, "
        "'{{{1,2},{3,4}},{{NULL,6},{7,8}}}'::FLOAT8[][][] AS f3")
    f1, f2, f3 = cursor.fetchone()
    assert f1 == [1, 2, 3, 4]
    assert f2 == [[1, 2, 3], [4, 5, 6]]
    assert f3 == [[[1, 2], [3, 4]], [[None, 6], [7, 8]]]


def test_int_array_roundtrip(cursor):
    # send small int array, should be sent as INT2[]
    retval = tuple(cursor.execute("SELECT %s as f1", ([1, 2, 3],)))
    assert retval[0][0], [1, 2, 3]
    column_name, column_typeoid = cursor.description[0][0:2]
    assert column_typeoid == 1005, "type should be INT2[]"

    # test multi-dimensional array, should be sent as INT2[]
    retval = tuple(cursor.execute("SELECT %s as f1", ([[1, 2], [3, 4]],)))
    assert retval[0][0] == [[1, 2], [3, 4]]

    column_name, column_typeoid = cursor.description[0][0:2]
    assert column_typeoid == 1005, "type should be INT2[]"

    # a larger value should kick it up to INT4[]...
    cursor.execute("SELECT %s as f1 -- integer[]", ([70000, 2, 3],))
    retval = cursor.fetchall()
    assert retval[0][0] == [70000, 2, 3]
    column_name, column_typeoid = cursor.description[0][0:2]
    assert column_typeoid == 1007, "type should be INT4[]"

    # a much larger value should kick it up to INT8[]...
    cursor.execute("SELECT %s as f1 -- bigint[]", ([7000000000, 2, 3],))
    retval = cursor.fetchall()
    assert retval[0][0] == [7000000000, 2, 3], "retrieved value match failed"
    column_name, column_typeoid = cursor.description[0][0:2]
    assert column_typeoid == 1016, "type should be INT8[]"


def test_int_array_with_null_roundtrip(cursor):
    retval = tuple(cursor.execute("SELECT %s as f1", ([1, None, 3],)))
    assert retval[0][0] == [1, None, 3]


def test_float_array_roundtrip(cursor):
    retval = tuple(cursor.execute("SELECT %s as f1", ([1.1, 2.2, 3.3],)))
    assert retval[0][0] == [1.1, 2.2, 3.3]


def test_bool_array_roundtrip(cursor):
    retval = tuple(cursor.execute("SELECT %s as f1", ([True, False, None],)))
    assert retval[0][0] == [True, False, None]


def test_string_array_out(cursor):
    cursor.execute("SELECT '{a,b,c}'::TEXT[] AS f1")
    assert cursor.fetchone()[0] == ["a", "b", "c"]
    cursor.execute("SELECT '{a,b,c}'::CHAR[] AS f1")
    assert cursor.fetchone()[0] == ["a", "b", "c"]
    cursor.execute("SELECT '{a,b,c}'::VARCHAR[] AS f1")
    assert cursor.fetchone()[0] == ["a", "b", "c"]
    cursor.execute("SELECT '{a,b,c}'::CSTRING[] AS f1")
    assert cursor.fetchone()[0] == ["a", "b", "c"]
    cursor.execute("SELECT '{a,b,c}'::NAME[] AS f1")
    assert cursor.fetchone()[0] == ["a", "b", "c"]
    cursor.execute("SELECT '{}'::text[];")
    assert cursor.fetchone()[0] == []
    cursor.execute("SELECT '{NULL,\"NULL\",NULL,\"\"}'::text[];")
    assert cursor.fetchone()[0] == [None, 'NULL', None, ""]


def test_numeric_array_out(cursor):
    cursor.execute("SELECT '{1.1,2.2,3.3}'::numeric[] AS f1")
    assert cursor.fetchone()[0] == [
        decimal.Decimal("1.1"), decimal.Decimal("2.2"), decimal.Decimal("3.3")]


def test_numeric_array_roundtrip(cursor):
    v = [decimal.Decimal("1.1"), None, decimal.Decimal("3.3")]
    retval = tuple(cursor.execute("SELECT %s as f1", (v,)))
    assert retval[0][0] == v


def test_string_array_roundtrip(cursor):
    v = [
        "Hello!", "World!", "abcdefghijklmnopqrstuvwxyz", "",
        "A bunch of random characters:",
        " ~!@#$%^&*()_+`1234567890-=[]\\{}|{;':\",./<>?\t", None]
    retval = tuple(cursor.execute("SELECT %s as f1", (v,)))
    assert retval[0][0] == v


def test_empty_array(cursor):
    v = []
    retval = tuple(cursor.execute("SELECT %s as f1", (v,)))
    assert retval[0][0] == v


def test_array_content_not_supported(con):
    class Kajigger(object):
        pass
    with pytest.raises(pg8000.ArrayContentNotSupportedError):
        con.array_inspect([[Kajigger()], [None], [None]])


def test_array_dimensions(con):
    for arr in (
            [1, [2]], [[1], [2], [3, 4]],
            [[[1]], [[2]], [[3, 4]]],
            [[[1]], [[2]], [[3, 4]]],
            [[[[1]]], [[[2]]], [[[3, 4]]]],
            [[1, 2, 3], [4, [5], 6]]):

        arr_send = con.array_inspect(arr)[2]
        with pytest.raises(pg8000.ArrayDimensionsNotConsistentError):
            arr_send(arr)


def test_array_homogenous(con):
    arr = [[[1]], [[2]], [[3.1]]]
    arr_send = con.array_inspect(arr)[2]
    with pytest.raises(pg8000.ArrayContentNotHomogenousError):
        arr_send(arr)


def test_array_inspect(con):
    con.array_inspect([1, 2, 3])
    con.array_inspect([[1], [2], [3]])
    con.array_inspect([[[1]], [[2]], [[3]]])


def test_macaddr(cursor):
    retval = tuple(cursor.execute("SELECT macaddr '08002b:010203'"))
    assert retval[0][0] == "08:00:2b:01:02:03"


def test_tsvector_roundtrip(cursor):
    cursor.execute(
        "SELECT cast(%s as tsvector)",
        ('a fat cat sat on a mat and ate a fat rat',))
    retval = cursor.fetchall()
    assert retval[0][0] == "'a' 'and' 'ate' 'cat' 'fat' 'mat' 'on' 'rat' 'sat'"


def test_hstore_roundtrip(cursor):
    val = '"a"=>"1"'
    retval = tuple(cursor.execute("SELECT cast(%s as hstore)", (val,)))
    assert retval[0][0] == val


def test_json_roundtrip(cursor):
    val = {'name': 'Apollo 11 Cave', 'zebra': True, 'age': 26.003}
    retval = tuple(cursor.execute("SELECT %s", (pg8000.PGJson(val),)))
    assert retval[0][0] == val


def test_jsonb_roundtrip(cursor):
    val = {'name': 'Apollo 11 Cave', 'zebra': True, 'age': 26.003}
    cursor.execute("SELECT cast(%s as jsonb)", (json.dumps(val),))
    retval = cursor.fetchall()
    assert retval[0][0] == val


def test_json_access_object(cursor):
    val = {'name': 'Apollo 11 Cave', 'zebra': True, 'age': 26.003}
    cursor.execute("SELECT cast(%s as json) -> %s", (json.dumps(val), 'name'))
    retval = cursor.fetchall()
    assert retval[0][0] == 'Apollo 11 Cave'


def test_jsonb_access_object(cursor):
    val = {'name': 'Apollo 11 Cave', 'zebra': True, 'age': 26.003}
    cursor.execute("SELECT cast(%s as jsonb) -> %s", (json.dumps(val), 'name'))
    retval = cursor.fetchall()
    assert retval[0][0] == 'Apollo 11 Cave'


def test_json_access_array(cursor):
    val = [-1, -2, -3, -4, -5]
    cursor.execute("SELECT cast(%s as json) -> %s", (json.dumps(val), 2))
    retval = cursor.fetchall()
    assert retval[0][0] == -3


def test_jsonb_access_array(cursor):
    val = [-1, -2, -3, -4, -5]
    cursor.execute("SELECT cast(%s as jsonb) -> %s", (json.dumps(val), 2))
    retval = cursor.fetchall()
    assert retval[0][0] == -3


def test_jsonb_access_path(cursor):
    j = {
        "a": [1, 2, 3],
        "b": [4, 5, 6]}

    path = ['a', '2']

    retval = tuple(cursor.execute("SELECT %s #>> %s", [PGJsonb(j), path]))
    assert retval[0][0] == str(j[path[0]][int(path[1])])


def test_timestamp_send_float():
    assert b'A\xbe\x19\xcf\x80\x00\x00\x00' == \
        pg8000.core.timestamp_send_float(Datetime(2016, 1, 2, 0, 0))


def test_infinity_timestamp_roundtrip(cursor):
    v = 'infinity'
    retval = tuple(cursor.execute("SELECT cast(%s as timestamp) as f1", (v,)))
    assert retval[0][0] == v


def test_point_roundtrip(cursor):
    v = '(2.3,1)'
    retval = tuple(cursor.execute("SELECT cast(%s as point) as f1", (v,)))
    assert retval[0][0] == v


def test_array_dim_lengths():
    assert pg8000.core.array_dim_lengths([[4], [5]]) == [2, 1]



import unittest
from pg8000 import Interval


# Type conversion tests
class Tests(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testIntervalConstructor(self):
        i = Interval(days=1)
        self.assertEqual(i.months, 0)
        self.assertEqual(i.days, 1)
        self.assertEqual(i.microseconds, 0)

    def intervalRangeTest(self, parameter, in_range, out_of_range):
        for v in out_of_range:
            try:
                Interval(**{parameter: v})
                self.fail("expected OverflowError")
            except OverflowError:
                pass
        for v in in_range:
            Interval(**{parameter: v})

    def testIntervalDaysRange(self):
        out_of_range_days = (-2147483648, +2147483648,)
        in_range_days = (-2147483647, +2147483647,)
        self.intervalRangeTest("days", in_range_days, out_of_range_days)

    def testIntervalMonthsRange(self):
        out_of_range_months = (-2147483648, +2147483648,)
        in_range_months = (-2147483647, +2147483647,)
        self.intervalRangeTest("months", in_range_months, out_of_range_months)

    def testIntervalMicrosecondsRange(self):
        out_of_range_microseconds = (
            -9223372036854775808, +9223372036854775808,)
        in_range_microseconds = (
            -9223372036854775807, +9223372036854775807,)
        self.intervalRangeTest(
            "microseconds", in_range_microseconds, out_of_range_microseconds)


if __name__ == "__main__":
    unittest.main()



from pg8000.core import convert_paramstyle as convert


# Tests of the convert_paramstyle function.

def test_qmark():
    new_query, make_args = convert(
        "qmark", "SELECT ?, ?, \"field_?\" FROM t "
        "WHERE a='say ''what?''' AND b=? AND c=E'?\\'test\\'?'")
    expected = "SELECT $1, $2, \"field_?\" FROM t WHERE " \
        "a='say ''what?''' AND b=$3 AND c=E'?\\'test\\'?'"
    assert new_query == expected
    assert make_args((1, 2, 3)) == (1, 2, 3)


def test_qmark_2():
    new_query, make_args = convert(
        "qmark", "SELECT ?, ?, * FROM t WHERE a=? AND b='are you ''sure?'")
    expected = "SELECT $1, $2, * FROM t WHERE a=$3 AND b='are you ''sure?'"
    assert new_query == expected
    assert make_args((1, 2, 3)) == (1, 2, 3)


def test_numeric():
    new_query, make_args = convert(
        "numeric",
        "SELECT sum(x)::decimal(5, 2) :2, :1, * FROM t WHERE a=:3")
    expected = "SELECT sum(x)::decimal(5, 2) $2, $1, * FROM t WHERE a=$3"
    assert new_query == expected
    assert make_args((1, 2, 3)) == (1, 2, 3)


def test_numeric_default_parameter():
    new_query, make_args = convert("numeric", "make_interval(days := 10)")

    assert new_query == "make_interval(days := 10)"
    assert make_args((1, 2, 3)) == (1, 2, 3)


def test_named():
    new_query, make_args = convert(
        "named",
        "SELECT sum(x)::decimal(5, 2) :f_2, :f1 FROM t WHERE a=:f_2")
    expected = "SELECT sum(x)::decimal(5, 2) $1, $2 FROM t WHERE a=$1"
    assert new_query == expected
    assert make_args({"f_2": 1, "f1": 2}) == (1, 2)


def test_format():
    new_query, make_args = convert(
        "format", "SELECT %s, %s, \"f1_%%\", E'txt_%%' "
        "FROM t WHERE a=%s AND b='75%%' AND c = '%' -- Comment with %")
    expected = "SELECT $1, $2, \"f1_%%\", E'txt_%%' FROM t WHERE a=$3 AND " \
        "b='75%%' AND c = '%' -- Comment with %"
    assert new_query == expected
    assert make_args((1, 2, 3)) == (1, 2, 3)

    sql = r"""COMMENT ON TABLE test_schema.comment_test """ \
        r"""IS 'the test % '' " \ table comment'"""
    new_query, make_args = convert("format", sql)
    assert new_query == sql


def test_format_multiline():
    new_query, make_args = convert("format", "SELECT -- Comment\n%s FROM t")
    assert new_query == "SELECT -- Comment\n$1 FROM t"


def test_py_format():
    new_query, make_args = convert(
        "pyformat", "SELECT %(f2)s, %(f1)s, \"f1_%%\", E'txt_%%' "
        "FROM t WHERE a=%(f2)s AND b='75%%'")
    expected = "SELECT $1, $2, \"f1_%%\", E'txt_%%' FROM t WHERE a=$1 AND " \
        "b='75%%'"
    assert new_query == expected
    assert make_args({"f2": 1, "f1": 2, "f3": 3}) == (1, 2)

    # pyformat should support %s and an array, too:
    new_query, make_args = convert(
        "pyformat", "SELECT %s, %s, \"f1_%%\", E'txt_%%' "
        "FROM t WHERE a=%s AND b='75%%'")
    expected = "SELECT $1, $2, \"f1_%%\", E'txt_%%' FROM t WHERE a=$3 AND " \
        "b='75%%'"
    assert new_query, expected
    assert make_args((1, 2, 3)) == (1, 2, 3)



#!/usr/bin/env python

import versioneer
from setuptools import setup

long_description = """\

pg8000
------

pg8000 is a Pure-Python interface to the PostgreSQL database engine.  It is \
one of many PostgreSQL interfaces for the Python programming language. pg8000 \
is somewhat distinctive in that it is written entirely in Python and does not \
rely on any external libraries (such as a compiled python module, or \
PostgreSQL's libpq library). pg8000 supports the standard Python DB-API \
version 2.0.

pg8000's name comes from the belief that it is probably about the 8000th \
PostgreSQL interface for Python."""

cmdclass = dict(versioneer.get_cmdclass())
version = versioneer.get_version()

setup(
    name="pg8000",
    version=version,
    cmdclass=cmdclass,
    description="PostgreSQL interface library",
    long_description=long_description,
    author="Mathieu Fenniak",
    author_email="biziqe@mathieu.fenniak.net",
    url="https://github.com/tlocke/pg8000",
    license="BSD",
    python_requires='>=3.5',
    install_requires=['scramp==1.1.1'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: Jython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="postgresql dbapi",
    packages=("pg8000",)
)



import pytest
import time
import warnings
import pg8000
''' Python DB API 2.0 driver compliance unit test suite.

    This software is Public Domain and may be used without restrictions.

 "Now we have booze and barflies entering the discussion, plus rumours of
  DBAs on drugs... and I won't tell you what flashes through my mind each
  time I read the subject line with 'Anal Compliance' in it.  All around
  this is turning out to be a thoroughly unwholesome unit test."

    -- Ian Bicking
'''

__rcs_id__ = '$Id: dbapi20.py,v 1.10 2003/10/09 03:14:14 zenzen Exp $'
__version__ = '$Revision: 1.10 $'[11:-2]
__author__ = 'Stuart Bishop <zen@shangri-la.dropbear.id.au>'


# $Log: dbapi20.py,v $
# Revision 1.10  2003/10/09 03:14:14  zenzen
# Add test for DB API 2.0 optional extension, where database exceptions
# are exposed as attributes on the Connection object.
#
# Revision 1.9  2003/08/13 01:16:36  zenzen
# Minor tweak from Stefan Fleiter
#
# Revision 1.8  2003/04/10 00:13:25  zenzen
# Changes, as per suggestions by M.-A. Lemburg
# - Add a table prefix, to ensure namespace collisions can always be avoided
#
# Revision 1.7  2003/02/26 23:33:37  zenzen
# Break out DDL into helper functions, as per request by David Rushby
#
# Revision 1.6  2003/02/21 03:04:33  zenzen
# Stuff from Henrik Ekelund:
#     added test_None
#     added test_nextset & hooks
#
# Revision 1.5  2003/02/17 22:08:43  zenzen
# Implement suggestions and code from Henrik Eklund - test that
# cursor.arraysize defaults to 1 & generic cursor.callproc test added
#
# Revision 1.4  2003/02/15 00:16:33  zenzen
# Changes, as per suggestions and bug reports by M.-A. Lemburg,
# Matthew T. Kromer, Federico Di Gregorio and Daniel Dittmar
# - Class renamed
# - Now a subclass of TestCase, to avoid requiring the driver stub
#   to use multiple inheritance
# - Reversed the polarity of buggy test in test_description
# - Test exception heirarchy correctly
# - self.populate is now self._populate(), so if a driver stub
#   overrides self.ddl1 this change propogates
# - VARCHAR columns now have a width, which will hopefully make the
#   DDL even more portible (this will be reversed if it causes more problems)
# - cursor.rowcount being checked after various execute and fetchXXX methods
# - Check for fetchall and fetchmany returning empty lists after results
#   are exhausted (already checking for empty lists if select retrieved
#   nothing
# - Fix bugs in test_setoutputsize_basic and test_setinputsizes
#


''' Test a database self.driver for DB API 2.0 compatibility.
    This implementation tests Gadfly, but the TestCase
    is structured so that other self.drivers can subclass this
    test case to ensure compiliance with the DB-API. It is
    expected that this TestCase may be expanded in the future
    if ambiguities or edge conditions are discovered.

    The 'Optional Extensions' are not yet being tested.

    self.drivers should subclass this test, overriding setUp, tearDown,
    self.driver, connect_args and connect_kw_args. Class specification
    should be as follows:

    import dbapi20
    class mytest(dbapi20.DatabaseAPI20Test):
       [...]

    Don't 'import DatabaseAPI20Test from dbapi20', or you will
    confuse the unit tester - just 'import dbapi20'.
'''

# The self.driver module. This should be the module where the 'connect'
# method is to be found
driver = pg8000
table_prefix = 'dbapi20test_'  # If you need to specify a prefix for tables

ddl1 = 'create table %sbooze (name varchar(20))' % table_prefix
ddl2 = 'create table %sbarflys (name varchar(20))' % table_prefix
xddl1 = 'drop table %sbooze' % table_prefix
xddl2 = 'drop table %sbarflys' % table_prefix

# Name of stored procedure to convert
# string->lowercase
lowerfunc = 'lower'


# Some drivers may need to override these helpers, for example adding
# a 'commit' after the execute.
def executeDDL1(cursor):
    cursor.execute(ddl1)


def executeDDL2(cursor):
    cursor.execute(ddl2)


@pytest.fixture
def db(request, con):
    def fin():
        with con.cursor() as cur:
            for ddl in (xddl1, xddl2):
                try:
                    cur.execute(ddl)
                    con.commit()
                except driver.Error:
                    # Assume table didn't exist. Other tests will check if
                    # execute is busted.
                    pass

    request.addfinalizer(fin)
    return con


def test_apilevel():
    # Must exist
    apilevel = driver.apilevel

    # Must equal 2.0
    assert apilevel == '2.0'


def test_threadsafety():
    try:
        # Must exist
        threadsafety = driver.threadsafety
        # Must be a valid value
        assert threadsafety in (0, 1, 2, 3)
    except AttributeError:
        assert False, "Driver doesn't define threadsafety"


def test_paramstyle():
    try:
        # Must exist
        paramstyle = driver.paramstyle
        # Must be a valid value
        assert paramstyle in (
            'qmark', 'numeric', 'named', 'format', 'pyformat')
    except AttributeError:
        assert False, "Driver doesn't define paramstyle"


def test_Exceptions():
    # Make sure required exceptions exist, and are in the
    # defined heirarchy.
    assert issubclass(driver.Warning, Exception)
    assert issubclass(driver.Error, Exception)
    assert issubclass(driver.InterfaceError, driver.Error)
    assert issubclass(driver.DatabaseError, driver.Error)
    assert issubclass(driver.OperationalError, driver.Error)
    assert issubclass(driver.IntegrityError, driver.Error)
    assert issubclass(driver.InternalError, driver.Error)
    assert issubclass(driver.ProgrammingError, driver.Error)
    assert issubclass(driver.NotSupportedError, driver.Error)


def test_ExceptionsAsConnectionAttributes(con):
    # OPTIONAL EXTENSION
    # Test for the optional DB API 2.0 extension, where the exceptions
    # are exposed as attributes on the Connection object
    # I figure this optional extension will be implemented by any
    # driver author who is using this test suite, so it is enabled
    # by default.
    warnings.simplefilter("ignore")
    drv = driver
    assert con.Warning is drv.Warning
    assert con.Error is drv.Error
    assert con.InterfaceError is drv.InterfaceError
    assert con.DatabaseError is drv.DatabaseError
    assert con.OperationalError is drv.OperationalError
    assert con.IntegrityError is drv.IntegrityError
    assert con.InternalError is drv.InternalError
    assert con.ProgrammingError is drv.ProgrammingError
    assert con.NotSupportedError is drv.NotSupportedError
    warnings.resetwarnings()


def test_commit(con):
    # Commit must work, even if it doesn't do anything
    con.commit()


def test_rollback(con):
    # If rollback is defined, it should either work or throw
    # the documented exception
    if hasattr(con, 'rollback'):
        try:
            con.rollback()
        except driver.NotSupportedError:
            pass


def test_cursor(con):
    con.cursor()


def test_cursor_isolation(con):
    # Make sure cursors created from the same connection have
    # the documented transaction isolation level
    cur1 = con.cursor()
    cur2 = con.cursor()
    executeDDL1(cur1)
    cur1.execute(
        "insert into %sbooze values ('Victoria Bitter')" % (table_prefix))
    cur2.execute("select name from %sbooze" % table_prefix)
    booze = cur2.fetchall()
    assert len(booze) == 1
    assert len(booze[0]) == 1
    assert booze[0][0] == 'Victoria Bitter'


def test_description(con):
    cur = con.cursor()
    executeDDL1(cur)
    assert cur.description is None, \
        'cursor.description should be none after executing a ' \
        'statement that can return no rows (such as DDL)'
    cur.execute('select name from %sbooze' % table_prefix)
    assert len(cur.description) == 1, \
        'cursor.description describes too many columns'
    assert len(cur.description[0]) == 7, \
        'cursor.description[x] tuples must have 7 elements'
    assert cur.description[0][0].lower() == b'name', \
        'cursor.description[x][0] must return column name'
    assert cur.description[0][1] == driver.STRING, \
        'cursor.description[x][1] must return column type. Got %r' \
        % cur.description[0][1]

    # Make sure self.description gets reset
    executeDDL2(cur)
    assert cur.description is None, \
        'cursor.description not being set to None when executing ' \
        'no-result statements (eg. DDL)'


def test_rowcount(cursor):
    executeDDL1(cursor)
    assert cursor.rowcount == -1, \
        'cursor.rowcount should be -1 after executing no-result ' \
        'statements'
    cursor.execute(
        "insert into %sbooze values ('Victoria Bitter')" % (table_prefix))
    assert cursor.rowcount in (-1, 1), \
        'cursor.rowcount should == number or rows inserted, or ' \
        'set to -1 after executing an insert statement'
    cursor.execute("select name from %sbooze" % table_prefix)
    assert cursor.rowcount in (-1, 1), \
        'cursor.rowcount should == number of rows returned, or ' \
        'set to -1 after executing a select statement'
    executeDDL2(cursor)
    assert cursor.rowcount == -1, \
        'cursor.rowcount not being reset to -1 after executing ' \
        'no-result statements'


lower_func = 'lower'


def test_callproc(cursor):
    if lower_func and hasattr(cursor, 'callproc'):
        r = cursor.callproc(lower_func, ('FOO',))
        assert len(r) == 1
        assert r[0] == 'FOO'
        r = cursor.fetchall()
        assert len(r) == 1, 'callproc produced no result set'
        assert len(r[0]) == 1, 'callproc produced invalid result set'
        assert r[0][0] == 'foo', 'callproc produced invalid results'


def test_close(con):
    cur = con.cursor()
    con.close()

    # cursor.execute should raise an Error if called after connection
    # closed
    with pytest.raises(driver.Error):
        executeDDL1(cur)

    # connection.commit should raise an Error if called after connection'
    # closed.'
    with pytest.raises(driver.Error):
        con.commit()

    # connection.close should raise an Error if called more than once
    with pytest.raises(driver.Error):
        con.close()


def test_execute(con):
    cur = con.cursor()
    _paraminsert(cur)


def _paraminsert(cur):
    executeDDL1(cur)
    cur.execute(
        "insert into %sbooze values ('Victoria Bitter')" % (table_prefix))
    assert cur.rowcount in (-1, 1)

    if driver.paramstyle == 'qmark':
        cur.execute(
            'insert into %sbooze values (?)' % table_prefix, ("Cooper's",))
    elif driver.paramstyle == 'numeric':
        cur.execute(
            'insert into %sbooze values (:1)' % table_prefix, ("Cooper's",))
    elif driver.paramstyle == 'named':
        cur.execute(
            'insert into %sbooze values (:beer)' % table_prefix,
            {'beer': "Cooper's"})
    elif driver.paramstyle == 'format':
        cur.execute(
            'insert into %sbooze values (%%s)' % table_prefix, ("Cooper's",))
    elif driver.paramstyle == 'pyformat':
        cur.execute(
            'insert into %sbooze values (%%(beer)s)' % table_prefix,
            {'beer': "Cooper's"})
    else:
        assert False, 'Invalid paramstyle'

    assert cur.rowcount in (-1, 1)

    cur.execute('select name from %sbooze' % table_prefix)
    res = cur.fetchall()
    assert len(res) == 2, 'cursor.fetchall returned too few rows'
    beers = [res[0][0], res[1][0]]
    beers.sort()
    assert beers[0] == "Cooper's", \
        'cursor.fetchall retrieved incorrect data, or data inserted ' \
        'incorrectly'
    assert beers[1] == "Victoria Bitter", \
        'cursor.fetchall retrieved incorrect data, or data inserted ' \
        'incorrectly'


def test_executemany(cursor):
    executeDDL1(cursor)
    largs = [("Cooper's",), ("Boag's",)]
    margs = [{'beer': "Cooper's"}, {'beer': "Boag's"}]
    if driver.paramstyle == 'qmark':
        cursor.executemany(
            'insert into %sbooze values (?)' % table_prefix, largs)
    elif driver.paramstyle == 'numeric':
        cursor.executemany(
            'insert into %sbooze values (:1)' % table_prefix, largs)
    elif driver.paramstyle == 'named':
        cursor.executemany(
            'insert into %sbooze values (:beer)' % table_prefix, margs)
    elif driver.paramstyle == 'format':
        cursor.executemany(
            'insert into %sbooze values (%%s)' % table_prefix, largs)
    elif driver.paramstyle == 'pyformat':
        cursor.executemany(
            'insert into %sbooze values (%%(beer)s)' % (table_prefix), margs)
    else:
        assert False, 'Unknown paramstyle'

    assert cursor.rowcount in (-1, 2), \
        'insert using cursor.executemany set cursor.rowcount to ' \
        'incorrect value %r' % cursor.rowcount

    cursor.execute('select name from %sbooze' % table_prefix)
    res = cursor.fetchall()
    assert len(res) == 2, 'cursor.fetchall retrieved incorrect number of rows'
    beers = [res[0][0], res[1][0]]
    beers.sort()
    assert beers[0] == "Boag's", 'incorrect data retrieved'
    assert beers[1] == "Cooper's", 'incorrect data retrieved'


def test_fetchone(cursor):
    # cursor.fetchone should raise an Error if called before
    # executing a select-type query
    with pytest.raises(driver.Error):
        cursor.fetchone()

    # cursor.fetchone should raise an Error if called after
    # executing a query that cannnot return rows
    executeDDL1(cursor)
    with pytest.raises(driver.Error):
        cursor.fetchone()

    cursor.execute('select name from %sbooze' % table_prefix)
    assert cursor.fetchone() is None, \
        'cursor.fetchone should return None if a query retrieves ' \
        'no rows'
    assert cursor.rowcount in (-1, 0)

    # cursor.fetchone should raise an Error if called after
    # executing a query that cannnot return rows
    cursor.execute(
        "insert into %sbooze values ('Victoria Bitter')" % (table_prefix))
    with pytest.raises(driver.Error):
        cursor.fetchone()

    cursor.execute('select name from %sbooze' % table_prefix)
    r = cursor.fetchone()
    assert len(r) == 1, 'cursor.fetchone should have retrieved a single row'
    assert r[0] == 'Victoria Bitter', \
        'cursor.fetchone retrieved incorrect data'
    assert cursor.fetchone() is None, \
        'cursor.fetchone should return None if no more rows available'
    assert cursor.rowcount in (-1, 1)


samples = [
    'Carlton Cold',
    'Carlton Draft',
    'Mountain Goat',
    'Redback',
    'Victoria Bitter',
    'XXXX'
]


def _populate():
    ''' Return a list of sql commands to setup the DB for the fetch
        tests.
    '''
    populate = [
        "insert into %sbooze values ('%s')" % (table_prefix, s)
        for s in samples]
    return populate


def test_fetchmany(cursor):
    # cursor.fetchmany should raise an Error if called without
    # issuing a query
    with pytest.raises(driver.Error):
        cursor.fetchmany(4)

    executeDDL1(cursor)
    for sql in _populate():
        cursor.execute(sql)

    cursor.execute('select name from %sbooze' % table_prefix)
    r = cursor.fetchmany()
    assert len(r) == 1, \
        'cursor.fetchmany retrieved incorrect number of rows, ' \
        'default of arraysize is one.'
    cursor.arraysize = 10
    r = cursor.fetchmany(3)  # Should get 3 rows
    assert len(r) == 3, 'cursor.fetchmany retrieved incorrect number of rows'
    r = cursor.fetchmany(4)  # Should get 2 more
    assert len(r) == 2, 'cursor.fetchmany retrieved incorrect number of rows'
    r = cursor.fetchmany(4)  # Should be an empty sequence
    assert len(r) == 0, \
        'cursor.fetchmany should return an empty sequence after ' \
        'results are exhausted'
    assert cursor.rowcount in (-1, 6)

    # Same as above, using cursor.arraysize
    cursor.arraysize = 4
    cursor.execute('select name from %sbooze' % table_prefix)
    r = cursor.fetchmany()  # Should get 4 rows
    assert len(r) == 4, 'cursor.arraysize not being honoured by fetchmany'
    r = cursor.fetchmany()  # Should get 2 more
    assert len(r) == 2
    r = cursor.fetchmany()  # Should be an empty sequence
    assert len(r) == 0
    assert cursor.rowcount in (-1, 6)

    cursor.arraysize = 6
    cursor.execute('select name from %sbooze' % table_prefix)
    rows = cursor.fetchmany()  # Should get all rows
    assert cursor.rowcount in (-1, 6)
    assert len(rows) == 6
    assert len(rows) == 6
    rows = [row[0] for row in rows]
    rows.sort()

    # Make sure we get the right data back out
    for i in range(0, 6):
        assert rows[i] == samples[i], \
            'incorrect data retrieved by cursor.fetchmany'

    rows = cursor.fetchmany()  # Should return an empty list
    assert len(rows) == 0, \
        'cursor.fetchmany should return an empty sequence if ' \
        'called after the whole result set has been fetched'
    assert cursor.rowcount in (-1, 6)

    executeDDL2(cursor)
    cursor.execute('select name from %sbarflys' % table_prefix)
    r = cursor.fetchmany()  # Should get empty sequence
    assert len(r) == 0, \
        'cursor.fetchmany should return an empty sequence if ' \
        'query retrieved no rows'
    assert cursor.rowcount in (-1, 0)


def test_fetchall(cursor):
    # cursor.fetchall should raise an Error if called
    # without executing a query that may return rows (such
    # as a select)
    with pytest.raises(driver.Error):
        cursor.fetchall()

    executeDDL1(cursor)
    for sql in _populate():
        cursor.execute(sql)

    # cursor.fetchall should raise an Error if called
    # after executing a a statement that cannot return rows
    with pytest.raises(driver.Error):
        cursor.fetchall()

    cursor.execute('select name from %sbooze' % table_prefix)
    rows = cursor.fetchall()
    assert cursor.rowcount in (-1, len(samples))
    assert len(rows) == len(samples), \
        'cursor.fetchall did not retrieve all rows'
    rows = [r[0] for r in rows]
    rows.sort()
    for i in range(0, len(samples)):
        assert rows[i] == samples[i], \
            'cursor.fetchall retrieved incorrect rows'
    rows = cursor.fetchall()
    assert len(rows) == 0, \
        'cursor.fetchall should return an empty list if called ' \
        'after the whole result set has been fetched'
    assert cursor.rowcount in (-1, len(samples))

    executeDDL2(cursor)
    cursor.execute('select name from %sbarflys' % table_prefix)
    rows = cursor.fetchall()
    assert cursor.rowcount in (-1, 0)
    assert len(rows) == 0, \
        'cursor.fetchall should return an empty list if ' \
        'a select query returns no rows'


def test_mixedfetch(cursor):
    executeDDL1(cursor)
    for sql in _populate():
        cursor.execute(sql)

    cursor.execute('select name from %sbooze' % table_prefix)
    rows1 = cursor.fetchone()
    rows23 = cursor.fetchmany(2)
    rows4 = cursor.fetchone()
    rows56 = cursor.fetchall()
    assert cursor.rowcount in (-1, 6)
    assert len(rows23) == 2, 'fetchmany returned incorrect number of rows'
    assert len(rows56) == 2, 'fetchall returned incorrect number of rows'

    rows = [rows1[0]]
    rows.extend([rows23[0][0], rows23[1][0]])
    rows.append(rows4[0])
    rows.extend([rows56[0][0], rows56[1][0]])
    rows.sort()
    for i in range(0, len(samples)):
        assert rows[i] == samples[i], 'incorrect data retrieved or inserted'


def help_nextset_setUp(cur):
    ''' Should create a procedure called deleteme
        that returns two result sets, first the
        number of rows in booze then "name from booze"
    '''
    raise NotImplementedError('Helper not implemented')


def help_nextset_tearDown(cur):
    'If cleaning up is needed after nextSetTest'
    raise NotImplementedError('Helper not implemented')


def test_nextset(cursor):
    if not hasattr(cursor, 'nextset'):
        return

    try:
        executeDDL1(cursor)
        sql = _populate()
        for sql in _populate():
            cursor.execute(sql)

        help_nextset_setUp(cursor)

        cursor.callproc('deleteme')
        numberofrows = cursor.fetchone()
        assert numberofrows[0] == len(samples)
        assert cursor.nextset()
        names = cursor.fetchall()
        assert len(names) == len(samples)
        s = cursor.nextset()
        assert s is None, 'No more return sets, should return None'
    finally:
        help_nextset_tearDown(cursor)


def test_arraysize(cursor):
    # Not much here - rest of the tests for this are in test_fetchmany
    assert hasattr(cursor, 'arraysize'), 'cursor.arraysize must be defined'


def test_setinputsizes(cursor):
    cursor.setinputsizes((25,))
    _paraminsert(cursor)  # Make sure cursor still works


def test_setoutputsize_basic(cursor):
    # Basic test is to make sure setoutputsize doesn't blow up
    cursor.setoutputsize(1000)
    cursor.setoutputsize(2000, 0)
    _paraminsert(cursor)  # Make sure the cursor still works


def test_None(cursor):
    executeDDL1(cursor)
    cursor.execute('insert into %sbooze values (NULL)' % table_prefix)
    cursor.execute('select name from %sbooze' % table_prefix)
    r = cursor.fetchall()
    assert len(r) == 1
    assert len(r[0]) == 1
    assert r[0][0] is None, 'NULL value not returned as None'


def test_Date():
    driver.Date(2002, 12, 25)
    driver.DateFromTicks(time.mktime((2002, 12, 25, 0, 0, 0, 0, 0, 0)))
    # Can we assume this? API doesn't specify, but it seems implied
    # self.assertEqual(str(d1),str(d2))


def test_Time():
    driver.Time(13, 45, 30)
    driver.TimeFromTicks(time.mktime((2001, 1, 1, 13, 45, 30, 0, 0, 0)))
    # Can we assume this? API doesn't specify, but it seems implied
    # self.assertEqual(str(t1),str(t2))


def test_Timestamp():
    driver.Timestamp(2002, 12, 25, 13, 45, 30)
    driver.TimestampFromTicks(time.mktime((2002, 12, 25, 13, 45, 30, 0, 0, 0)))
    # Can we assume this? API doesn't specify, but it seems implied
    # self.assertEqual(str(t1),str(t2))


def test_Binary():
    driver.Binary(b'Something')
    driver.Binary(b'')


def test_STRING():
    assert hasattr(driver, 'STRING'), 'module.STRING must be defined'


def test_BINARY():
    assert hasattr(driver, 'BINARY'), 'module.BINARY must be defined.'


def test_NUMBER():
    assert hasattr(driver, 'NUMBER'), 'module.NUMBER must be defined.'


def test_DATETIME():
    assert hasattr(driver, 'DATETIME'), 'module.DATETIME must be defined.'


def test_ROWID():
    assert hasattr(driver, 'ROWID'), 'module.ROWID must be defined.'
